/*
 * File : list.h
 *
 * Description : This file contains the listMember and list class
 * declarations. Also defined are macros DefineListMember and
 * DefineList.
 *
 * Portions of this work are derived from A C++ Toolkit, which is 
 * Copyright (C) 1991 by Jonathan S. Shapiro, and are used with
 * permission. A C++ Toolkit is published by Prentice Hall, Inc.
 *
 */

#ifndef _list_h_
#define _list_h_

#include <generic.h>


/*
 * Class : listMember
 *
 * Description : The listMember class provides the basic 
 * functionality for an object to be a part of a doubly-
 * linked list. Classes, instances of which need to be
 * part of a list, will inherit from listMember.
 *
 * Attributes :
 *	listMember :_next : Pointer to the next listMember in list.
 *	listMember :_prev : Pointer to the previous listMember in list.
 */

class listMember {
	friend class list;
	int	_marked;

 protected:

	listMember *_next;
	listMember *_prev;

 public:
	
	listMember()
		{ _next = _prev = 0; _marked = 0; }

	virtual ~listMember();
	listMember *Next()
		{ return _next; }

	listMember *Prev()
		{ return _prev; }

	int IsMarked() 
		{ return _marked; }

	void SetMarked()
		{ _marked = 1; }
};

/*
 * Class : list
 *
 * Description : The list class provides the functionality
 * to manipulate a doubly linked list of listMember objects.
 *
 * Attributes:
 *	_head : The head of the list.
 *	_tail : The tail of the list.
 */


class list {
 protected:
	listMember *_head;
	listMember *_tail;

 public:

	list() 
		{ _head = _tail = 0; }

	virtual ~list() {};

	listMember *Head()
		{ return _head; }
	listMember *Tail()
		{ return _tail; }

	list& Append(listMember *);
	list& Prepend(listMember *);
	list& Remove(listMember *);
	list& CleanUpMarked();
	int	  Length();

	listMember *GetNthMember(int n);

	list& Push(listMember *l)
		{ return Append(l); }

	listMember *Pop()
		{	listMember *l = _tail;
			Remove(_tail);
			return l;
		}	

	listMember *Top()
		{ return _tail; }
};

/*
 * Macro : DefineListMember
 *
 * Description : This macro defines a type-safe sub-class
 * of the listMember class.
 */

#define DefineListMember(TYPE)					\
												\
class name2(TYPE,Member) : public listMember {	\
 public:										\
	name2(TYPE,Member) () : listMember() {}		\
	virtual ~name2(TYPE,Member) () {}				\
												\
	TYPE *Next() {return (TYPE *)_next;}		\
	TYPE *Prev() {return (TYPE *)_prev;}		\
};	

/*
 * Macro : DefineList
 *
 * Description : This macro defines a type-safe sub-class
 * of the list class.
 */

#define DefineList(TYPE)					\
											\
class name2(TYPE,List) : public list {		\
 public:									\
	name2(TYPE,List) () : list() {}			\
	virtual ~name2(TYPE,List) () {} 		\
											\
	TYPE *Tail()							\
		{return (TYPE *) _tail; }			\
	TYPE *Head()							\
		{return (TYPE *) _head;}			\
	TYPE *Pop()								\
		{ return (TYPE *)list::Pop(); }		\
	TYPE *Top()								\
		{ return (TYPE *) _tail; }			\
	TYPE *GetNthMember(int n)		\
		{ return (TYPE *)list::GetNthMember(n); }		\
	name2(TYPE,List)& Append(name2(TYPE,Member) *l)	\
		{									\
			list::Append(l);				\
			return *this;					\
		}									\
	name2(TYPE,List)& Prepend(name2(TYPE,Member) *l)\
		{									\
			list::Prepend(l);				\
			return *this;					\
		}									\
	name2(TYPE,List)& Remove(name2(TYPE,Member) *l) \
		{									\
			list::Remove(l);				\
			return *this;					\
		}									\
	name2(TYPE,List)& Push(name2(TYPE,Member) *l) 	\
		{									\
			list::Push(l);					\
			return *this;					\
		}									\
};
#endif
